@@ -1,15 +1,18 @@
 # Licensed under the MIT License
 # https://github.com/craigahobbs/chisel/blob/master/LICENSE
 
+"""
+TODO
+"""
+
 from datetime import datetime, timedelta
 from http import HTTPStatus
+from io import BytesIO
 import logging
 import re
 from urllib.parse import quote, unquote
 
-from .app_defs import Environ, StartResponse
-from .url import encode_query_string
-from .util import JSONEncoder
+from .util import encode_query_string, JSONEncoder
 
 
 # Regular expression for matching URL arguments
@@ -19,7 +22,7 @@ RE_URL_ARG_ESC = re.compile(r'/\\{([A-Za-z]\w*)\\}')
 
 class Application:
     """
-    Chisel application
+    TODO
     """
 
     __slots__ = (
@@ -33,22 +36,33 @@ class Application:
     )
 
     def __init__(self):
+
+        #: TODO
         self.log_level = logging.WARNING
+
+        #: TODO
         self.log_format = '%(levelname)s [%(process)s / %(thread)s] %(message)s'
+
+        #: TODO
         self.pretty_output = False
+
+        #: TODO
         self.validate_output = True
+
+        #: TODO
         self.requests = {}
+
         self.__request_urls = {}
         self.__request_regex = []
 
     def add_request(self, request):
         """
-        Add a request object
+        TODO
         """
 
         # Duplicate request name?
         if request.name in self.requests:
-            raise ValueError('redefinition of request "{0}"'.format(request.name))
+            raise ValueError(f'redefinition of request "{request.name}"')
         self.requests[request.name] = request
 
         # Add the request URLs
@@ -61,12 +75,20 @@ class Application:
             else:
                 request_key = (method, url)
                 if request_key in self.__request_urls:
-                    raise ValueError('redefinition of request URL "{0}"'.format(url))
+                    raise ValueError(f'redefinition of request URL "{url}"')
                 self.__request_urls[request_key] = request
 
+    def add_requests(self, requests):
+        """
+        TODO
+        """
+
+        for request in requests:
+            self.add_request(request)
+
     def __call__(self, environ, start_response):
         """
-        Chisel application WSGI entry point
+        TODO
         """
 
         # Match the request by method and exact URL
@@ -112,7 +134,7 @@ class Application:
                     )
 
         # Create the request context
-        ctx = environ[Environ.CTX] = Context(self, environ, start_response, url_args)
+        ctx = environ[Context.ENVIRON_CTX] = Context(self, environ, start_response, url_args)
 
         # Request not found?
         if request is None:
@@ -134,7 +156,11 @@ class Application:
         return response
 
     def request(self, request_method, path_info, query_string='', wsgi_input=b'', environ=None):
-        request_environ = Environ.create(request_method, path_info, query_string, wsgi_input, environ=environ)
+        """
+        TODO
+        """
+
+        request_environ = Context.create_environ(request_method, path_info, query_string, wsgi_input, environ=environ)
         start_response = StartResponse()
         response = self(request_environ, start_response)
         return start_response.status, start_response.headers, b''.join(response)
@@ -142,19 +168,31 @@ class Application:
 
 class Context:
     """
-    Chisel request context
+    TODO
     """
 
     __slots__ = ('app', 'environ', '_start_response', 'url_args', 'log', 'headers')
 
+    #: TODO
+    ENVIRON_CTX = 'chisel.ctx'
+
     def __init__(self, app, environ=None, start_response=None, url_args=None):
+
+        #: TODO
         self.app = app
+
+        #: TODO
         self.environ = environ or {}
+
         self._start_response = start_response
+
+        #: TODO
         self.url_args = url_args
+
+        #: TODO
         self.headers = {}
 
-        # Create the logger
+        #: TODO
         self.log = logging.getLoggerClass()('')
         self.log.setLevel(app.log_level)
         wsgi_errors = environ.get('wsgi.errors') if environ else None
@@ -168,16 +206,38 @@ class Context:
             handler.setFormatter(logging.Formatter(app.log_format))
         self.log.addHandler(handler)
 
+    @staticmethod
+    def create_environ(request_method, path_info, query_string='', wsgi_input=b'', environ=None):
+        """
+        TODO
+        """
+
+        if environ is None:
+            environ = {}
+        environ.setdefault('REQUEST_METHOD', request_method)
+        environ.setdefault('PATH_INFO', path_info)
+        environ.setdefault('QUERY_STRING', query_string if isinstance(query_string, str) else encode_query_string(query_string))
+        environ.setdefault('SCRIPT_NAME', '')
+        environ.setdefault('SERVER_NAME', 'localhost')
+        environ.setdefault('SERVER_PORT', '80')
+        environ.setdefault('wsgi.input', BytesIO(wsgi_input))
+        environ.setdefault('wsgi.url_scheme', 'http')
+        return environ
+
     def start_response(self, status, headers):
+        """
+        TODO
+        """
+
         if not isinstance(status, str):
-            status = str(status.value) + ' ' + status.phrase
+            status = f'{status.value} {status.phrase}'
         for key, value in headers:
             self.add_header(key, value)
         self._start_response(status, sorted(self.headers.items()))
 
     def add_header(self, key, value):
         """
-        Add a response header
+        TODO
         """
 
         assert isinstance(key, str), 'header key must be of type str'
@@ -185,6 +245,10 @@ class Context:
         self.headers[key] = value
 
     def add_cache_headers(self, control=None, ttl_seconds=None, utcnow=None):
+        """
+        TODO
+        """
+
         if self.environ.get('REQUEST_METHOD') == 'GET':
             if control is None:
                 assert ttl_seconds is None
@@ -192,14 +256,14 @@ class Context:
             else:
                 assert control in ('public', 'private')
                 assert isinstance(ttl_seconds, int) and ttl_seconds > 0
-                self.add_header('Cache-Control', '{0},max-age={1}'.format(control, ttl_seconds))
+                self.add_header('Cache-Control', f'{control},max-age={ttl_seconds}')
                 if utcnow is None:
                     utcnow = datetime.utcnow()
                 self.add_header('Expires', (utcnow + timedelta(seconds=ttl_seconds)).strftime('%a, %d %b %Y %H:%M:%S GMT'))
 
     def response(self, status, content_type, content, headers=None):
         """
-        Send an HTTP response
+        TODO
         """
 
         assert not isinstance(content, (str, bytes)), 'response content cannot be of type str or bytes'
@@ -211,7 +275,7 @@ class Context:
 
     def response_text(self, status, text=None, content_type='text/plain', encoding='utf-8', headers=None):
         """
-        Send a plain-text response
+        TODO
         """
 
         if text is None:
@@ -223,7 +287,7 @@ class Context:
 
     def response_json(self, status, response, content_type='application/json', encoding='utf-8', headers=None, jsonp=None):
         """
-        Send a JSON response
+        TODO
         """
 
         encoder = JSONEncoder(
@@ -242,7 +306,7 @@ class Context:
 
     def reconstruct_url(self, path_info=None, query_string=None, relative=False):
         """
-        Reconstructs the request URL using the algorithm provided by PEP3333
+        TODO
         """
 
         environ = self.environ
@@ -279,3 +343,20 @@ class Context:
                     url += '?' + encode_query_string(query_string)
 
         return url
+
+
+class StartResponse:
+    """
+    TODO
+    """
+
+    __slots__ = ('status', 'headers')
+
+    def __init__(self):
+        self.status = None
+        self.headers = None
+
+    def __call__(self, status, headers):
+        assert self.status is None and self.headers is None
+        self.status = status
+        self.headers = headers
