@@ -1,6 +1,10 @@
 # Licensed under the MIT License
 # https://github.com/craigahobbs/chisel/blob/master/LICENSE
 
+"""
+TODO
+"""
+
 from datetime import date, datetime
 from decimal import Decimal
 from enum import Enum
@@ -8,110 +12,237 @@ from itertools import chain
 from math import isnan, isinf
 from uuid import UUID
 
-from .util import TZLOCAL, parse_iso8601_date, parse_iso8601_datetime
-
 
 # Validation mode
 class ValidationMode(Enum):
+    """
+    TODO
+    """
+
+    #: TODO
     DEFAULT = 0
+
+    #: TODO
     QUERY_STRING = 1
+
+    #: TODO
     JSON_INPUT = 2
 
+
 # Immutable validation modes
 IMMUTABLE_VALIDATION_MODES = (ValidationMode.DEFAULT,)
 
 
-# Type attribute exception
 class AttributeValidationError(Exception):
+    """
+    TODO
+    """
+
     __slots__ = ('attr',)
 
     def __init__(self, attr):
         super().__init__("Invalid attribute '" + attr + "'")
+
+        #: TODO
         self.attr = attr
 
 
-# Type validation exception
 class ValidationError(Exception):
+    """
+    TODO
+    """
+
     __slots__ = ('member',)
 
     def __init__(self, msg, member=None):
         super().__init__(msg)
+
+        #: TODO
         self.member = member
 
-    @classmethod
-    def _flatten_members(cls, members):
-        for submember in members:
-            if isinstance(submember, tuple):
-                yield from cls._flatten_members(submember)
-            else:
-                yield submember
 
-    @classmethod
-    def member_syntax(cls, members):
-        if members:
-            return ''.join((('.' + x) if isinstance(x, str) else ('[' + repr(x) + ']'))
-                           for x in cls._flatten_members(members)).lstrip('.')
-        return None
+def _flatten_members(members):
+    for submember in members:
+        if isinstance(submember, tuple):
+            yield from _flatten_members(submember)
+        else:
+            yield submember
 
-    @classmethod
-    def member_error(cls, type_, value, members, constraint_syntax=None):
-        member_syntax = cls.member_syntax(members)
-        msg = "Invalid value {0!r:.1000s} (type '{1}'){2}{3}{4}".format(
-            value,
-            value.__class__.__name__,
-            " for member '" + member_syntax + "'" if member_syntax else '',
-            ", expected type '" + type_.type_name + "'" if type_ else '',
-            ' [' + constraint_syntax + ']' if constraint_syntax else ''
+def _member_syntax(members):
+    if members:
+        return ''.join(((f'.{x}') if isinstance(x, str) else (f'[{x!r}]')) for x in _flatten_members(members)).lstrip('.')
+    return None
+
+def _member_error(type_, value, members, constraint_syntax=None):
+    member_syntax = _member_syntax(members)
+    member_part = f" for member '{member_syntax}'" if member_syntax else ''
+    type_part = f", expected type '{type_.type_name}'" if type_ else ''
+    constraint_part = f' [{constraint_syntax}]' if constraint_syntax else ''
+    msg = f"Invalid value {value!r:.1000s} (type '{value.__class__.__name__}'){member_part}{type_part}{constraint_part}"
+    return ValidationError(msg, member=member_syntax)
+
+
+def get_referenced_types(type_):
+    """
+    TODO
+    """
+
+    return sorted(_get_referenced_types(type_, set(), True), key=lambda type_: type_.type_name)
+
+def _get_referenced_types(type_, visited, top_level):
+    if isinstance(type_, TypeStruct):
+        if type_.type_name not in visited:
+            visited.add(type_.type_name)
+            if not top_level:
+                yield type_
+            for member in type_.members():
+                yield from _get_referenced_types(member.type, visited, False)
+    elif isinstance(type_, TypeEnum):
+        if type_.type_name not in visited:
+            visited.add(type_.type_name)
+            if not top_level:
+                yield type_
+    elif isinstance(type_, Typedef):
+        if type_.type_name not in visited:
+            visited.add(type_.type_name)
+            if not top_level:
+                yield type_
+            yield from _get_referenced_types(type_.type, visited, False)
+    elif isinstance(type_, TypeArray):
+        yield from _get_referenced_types(type_.type, visited, False)
+    elif isinstance(type_, TypeDict):
+        yield from _get_referenced_types(type_.type, visited, False)
+        yield from _get_referenced_types(type_.key_type, visited, False)
+    elif isinstance(type_, ActionModel):
+        yield from _get_referenced_types(type_.path_type, visited, True)
+        yield from _get_referenced_types(type_.query_type, visited, True)
+        yield from _get_referenced_types(type_.input_type, visited, True)
+        yield from _get_referenced_types(type_.output_type, visited, True)
+
+
+class ActionModel:
+    """
+    TODO
+    """
+
+    __slots__ = ('name', 'urls', 'path_type', 'query_type', 'input_type', 'output_type', 'error_type', 'doc', 'doc_group')
+
+    def __init__(self, name, doc=None, doc_group=None):
+
+        #: TODO
+        self.name = name
+
+        #: TODO
+        self.urls = []
+
+        #: TODO
+        self.path_type = TypeStruct(type_name=name + '_path')
+
+        #: TODO
+        self.query_type = TypeStruct(type_name=name + '_query')
+
+        #: TODO
+        self.input_type = TypeStruct(type_name=name + '_input')
+
+        #: TODO
+        self.output_type = TypeStruct(type_name=name + '_output')
+
+        #: TODO
+        self.error_type = TypeEnum(type_name=name + '_error')
+
+        #: TODO
+        self.doc = [] if doc is None else doc
+
+        #: TODO
+        self.doc_group = doc_group
+
+    def input_members(self, include_base_types=True):
+        """
+        TODO
+        """
+
+        yield from chain.from_iterable(
+            t.members(include_base_types=include_base_types) for t in (self.path_type, self.query_type, self.input_type)
         )
-        return ValidationError(msg, member=member_syntax)
 
 
 # Struct member attributes
 class StructMemberAttributes:
+    """
+    TODO
+    """
+
     __slots__ = ('op_eq', 'op_lt', 'op_lte', 'op_gt', 'op_gte',
                  'op_len_eq', 'op_len_lt', 'op_len_lte', 'op_len_gt', 'op_len_gte')
 
     def __init__(self, op_eq=None, op_lt=None, op_lte=None, op_gt=None, op_gte=None,
                  op_len_eq=None, op_len_lt=None, op_len_lte=None, op_len_gt=None, op_len_gte=None):
 
+        #: TODO
         self.op_eq = op_eq
+
+        #: TODO
         self.op_lt = op_lt
+
+        #: TODO
         self.op_lte = op_lte
+
+        #: TODO
         self.op_gt = op_gt
+
+        #: TODO
         self.op_gte = op_gte
+
+        #: TODO
         self.op_len_eq = op_len_eq
+
+        #: TODO
         self.op_len_lt = op_len_lt
+
+        #: TODO
         self.op_len_lte = op_len_lte
+
+        #: TODO
         self.op_len_gt = op_len_gt
+
+        #: TODO
         self.op_len_gte = op_len_gte
 
     @staticmethod
     def _format_float(value):
-        return '{0:.6f}'.format(value).rstrip('0').rstrip('.')
+        return f'{value:.6f}'.rstrip('0').rstrip('.')
 
     def validate(self, value, _member=()):
+        """
+        TODO
+        """
+
         if self.op_lt is not None and value >= self.op_lt:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='< ' + self._format_float(self.op_lt))
+            raise _member_error(None, value, _member, constraint_syntax='< ' + self._format_float(self.op_lt))
         if self.op_lte is not None and value > self.op_lte:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='<= ' + self._format_float(self.op_lte))
+            raise _member_error(None, value, _member, constraint_syntax='<= ' + self._format_float(self.op_lte))
         if self.op_gt is not None and value <= self.op_gt:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='> ' + self._format_float(self.op_gt))
+            raise _member_error(None, value, _member, constraint_syntax='> ' + self._format_float(self.op_gt))
         if self.op_gte is not None and value < self.op_gte:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='>= ' + self._format_float(self.op_gte))
+            raise _member_error(None, value, _member, constraint_syntax='>= ' + self._format_float(self.op_gte))
         if self.op_eq is not None and value != self.op_eq:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='== ' + self._format_float(self.op_eq))
+            raise _member_error(None, value, _member, constraint_syntax='== ' + self._format_float(self.op_eq))
         if self.op_len_lt is not None and len(value) >= self.op_len_lt:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='len < ' + self._format_float(self.op_len_lt))
+            raise _member_error(None, value, _member, constraint_syntax='len < ' + self._format_float(self.op_len_lt))
         if self.op_len_lte is not None and len(value) > self.op_len_lte:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='len <= ' + self._format_float(self.op_len_lte))
+            raise _member_error(None, value, _member, constraint_syntax='len <= ' + self._format_float(self.op_len_lte))
         if self.op_len_gt is not None and len(value) <= self.op_len_gt:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='len > ' + self._format_float(self.op_len_gt))
+            raise _member_error(None, value, _member, constraint_syntax='len > ' + self._format_float(self.op_len_gt))
         if self.op_len_gte is not None and len(value) < self.op_len_gte:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='len >= ' + self._format_float(self.op_len_gte))
+            raise _member_error(None, value, _member, constraint_syntax='len >= ' + self._format_float(self.op_len_gte))
         if self.op_len_eq is not None and len(value) != self.op_len_eq:
-            raise ValidationError.member_error(None, value, _member, constraint_syntax='len == ' + self._format_float(self.op_len_eq))
+            raise _member_error(None, value, _member, constraint_syntax='len == ' + self._format_float(self.op_len_eq))
 
     def validate_attr(self, allow_value=False, allow_length=False):
+        """
+        TODO
+        """
+
         if not allow_value:
             if self.op_lt is not None:
                 raise AttributeValidationError('< ' + self._format_float(self.op_lt))
@@ -136,70 +267,139 @@ class StructMemberAttributes:
                 raise AttributeValidationError('len == ' + self._format_float(self.op_len_eq))
 
 
-# Typedef type (type plus attributes)
 class Typedef:
+    """
+    TODO
+    """
+
     __slots__ = ('type_name', 'type', 'attr', 'doc')
 
     def __init__(self, type_, attr=None, type_name=None, doc=None):
+
+        #: TODO
         self.type_name = 'typedef' if type_name is None else type_name
+
+        #: TODO
         self.type = type_
+
+        #: TODO
         self.attr = attr
+
+        #: TODO
         self.doc = [] if doc is None else doc
 
     @staticmethod
     def base_type(type_):
+        """
+        TODO
+        """
+
         while isinstance(type_, Typedef):
             type_ = type_.type
         return type_
 
     def validate_attr(self, attr):
+        """
+        TODO
+        """
+
         self.type.validate_attr(attr)
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
+
         result = self.type.validate(value, mode, _member)
         if self.attr is not None:
             self.attr.validate(result, _member)
         return result
 
 
-# Struct member
 class StructMember:
+    """
+    TODO
+    """
+
     __slots__ = ('name', 'type', 'optional', 'nullable', 'attr', 'doc')
 
     def __init__(self, name, type_, optional=False, nullable=False, attr=None, doc=None):
+
+        #: TODO
         self.name = name
+
+        #: TODO
         self.type = type_
+
+        #: TODO
         self.optional = optional
+
+        #: TODO
         self.nullable = nullable
+
+        #: TODO
         self.attr = attr
+
+        #: TODO
         self.doc = [] if doc is None else doc
 
 
-# Struct type
 class TypeStruct:
+    """
+    TODO
+    """
+
     __slots__ = ('type_name', 'union', 'base_types', '_members', 'doc')
 
     def __init__(self, type_name=None, union=False, base_types=None, doc=None):
+
+        #: TODO
         self.type_name = ('union' if union else 'struct') if type_name is None else type_name
+
+        #: TODO
         self.union = union
+
+        #: TODO
         self.base_types = base_types
+
         self._members = []
+
+        #: TODO
         self.doc = [] if doc is None else doc
 
     def members(self, include_base_types=True):
-        return iter(self._members) if self.base_types is None or not include_base_types else \
-            chain(chain.from_iterable(Typedef.base_type(base_type).members() for base_type in self.base_types if base_type), self._members)
+        """
+        TODO
+        """
+
+        if include_base_types and self.base_types is not None:
+            return chain(
+                chain.from_iterable(Typedef.base_type(base_type).members() for base_type in self.base_types if base_type),
+                self._members
+            )
+        return iter(self._members)
 
     def add_member(self, name, type_, optional=False, nullable=False, attr=None, doc=None):
+        """
+        TODO
+        """
+
         member = StructMember(name, type_, optional or self.union, nullable, attr, doc)
         self._members.append(member)
         return member
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr()
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, dict):
@@ -207,12 +407,12 @@ class TypeStruct:
         elif mode == ValidationMode.QUERY_STRING and value == '':
             value_x = {}
         else:
-            raise ValidationError.member_error(self, value, _member)
+            raise _member_error(self, value, _member)
 
         # Valid union?
         if self.union:
             if len(value_x) != 1:
-                raise ValidationError.member_error(self, value, _member)
+                raise _member_error(self, value, _member)
 
         # Result a copy?
         value_copy = None if mode in IMMUTABLE_VALIDATION_MODES else {}
@@ -223,45 +423,60 @@ class TypeStruct:
             member_name = member.name
             if member_name not in value_x:
                 if not member.optional:
-                    raise ValidationError("Required member {0!r} missing".format(ValidationError.member_syntax((_member, member_name))))
-                continue
-            member_count += 1
-            member_value = value_x[member_name]
-            if member.nullable and (member_value is None or \
-                    (mode == ValidationMode.QUERY_STRING and not isinstance(member.type, _TypeString) and member_value == 'null')):
-                member_value_x = None
+                    raise ValidationError(f"Required member {_member_syntax((_member, member_name))!r} missing")
             else:
-                member_path = (_member, member_name)
-                member_value_x = member.type.validate(member_value, mode, member_path)
-                if member.attr is not None:
-                    member.attr.validate(member_value_x, member_path)
-            if value_copy is not None:
-                value_copy[member_name] = member_value_x
+                member_count += 1
+                member_value = value_x[member_name]
+                if member.nullable and (member_value is None or \
+                        (mode == ValidationMode.QUERY_STRING and not isinstance(member.type, TypeString) and member_value == 'null')):
+                    member_value_x = None
+                else:
+                    member_path = (_member, member_name)
+                    member_value_x = member.type.validate(member_value, mode, member_path)
+                    if member.attr is not None:
+                        member.attr.validate(member_value_x, member_path)
+                if value_copy is not None:
+                    value_copy[member_name] = member_value_x
 
         # Any unknown members?
         if member_count != len(value_x):
             member_set = {member.name for member in self.members()}
             unknown_value_names = [value_name for value_name in value_x.keys() if value_name not in member_set]
-            raise ValidationError("Unknown member {0!r:.100s}".format(ValidationError.member_syntax((_member, unknown_value_names[0]))))
+            raise ValidationError(f"Unknown member {_member_syntax((_member, unknown_value_names[0]))!r:.100s}")
 
         return value if value_copy is None else value_copy
 
 
-# Array type
 class TypeArray:
+    """
+    TODO
+    """
+
     __slots__ = ('type', 'attr')
 
+    #: TODO
     type_name = 'array'
 
     def __init__(self, type_, attr=None):
+
+        #: TODO
         self.type = type_
+
+        #: TODO
         self.attr = attr
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr(allow_length=True)
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, (list, tuple)):
@@ -269,7 +484,7 @@ class TypeArray:
         elif mode == ValidationMode.QUERY_STRING and value == '':
             value_x = []
         else:
-            raise ValidationError.member_error(self, value, _member)
+            raise _member_error(self, value, _member)
 
         # Result a copy?
         value_copy = None if mode in IMMUTABLE_VALIDATION_MODES else []
@@ -288,31 +503,58 @@ class TypeArray:
         return value if value_copy is None else value_copy
 
 
-# Dict type
 class TypeDict:
+    """
+    TODO
+    """
+
     __slots__ = ('type', 'attr', 'key_type', 'key_attr')
 
+    #: TODO
     type_name = 'dict'
 
     def __init__(self, type_, attr=None, key_type=None, key_attr=None):
+
+        #: TODO
         self.type = type_
+
+        #: TODO
         self.attr = attr
+
+        #: TODO
         self.key_type = key_type or TYPE_STRING
+
+        #: TODO
         self.key_attr = key_attr
 
     @staticmethod
     def valid_key_type(key_type):
+        """
+        TODO
+        """
+
         key_type_base = Typedef.base_type(key_type)
-        return isinstance(key_type_base, (_TypeString, TypeEnum))
+        return isinstance(key_type_base, (TypeString, TypeEnum))
 
     def has_default_key_type(self):
-        return isinstance(self.key_type, _TypeString)
+        """
+        TODO
+        """
+
+        return isinstance(self.key_type, TypeString)
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr(allow_length=True)
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, dict):
@@ -320,7 +562,7 @@ class TypeDict:
         elif mode == ValidationMode.QUERY_STRING and value == '':
             value_x = {}
         else:
-            raise ValidationError.member_error(self, value, _member)
+            raise _member_error(self, value, _member)
 
         # Result a copy?
         value_copy = None if mode in IMMUTABLE_VALIDATION_MODES else {}
@@ -346,12 +588,19 @@ class TypeDict:
         return value if value_copy is None else value_copy
 
 
-# Enumeration type
 class EnumValue:
+    """
+    TODO
+    """
+
     __slots__ = ('value', 'doc')
 
     def __init__(self, valueString, doc=None):
+
+        #: TODO
         self.value = valueString
+
+        #: TODO
         self.doc = [] if doc is None else doc
 
     def __eq__(self, other):
@@ -359,68 +608,118 @@ class EnumValue:
 
 
 class TypeEnum:
+    """
+    TODO
+    """
+
     __slots__ = ('type_name', 'base_types', '_values', 'doc')
 
     def __init__(self, type_name='enum', base_types=None, doc=None):
+
+        #: TODO
         self.type_name = type_name
+
+        #: TODO
         self.base_types = base_types
+
         self._values = []
+
+        #: TODO
         self.doc = [] if doc is None else doc
 
     def values(self, include_base_types=True):
-        return iter(self._values) if self.base_types is None or not include_base_types else \
-            chain(chain.from_iterable(Typedef.base_type(base_type).values() for base_type in self.base_types if base_type), self._values)
+        """
+        TODO
+        """
+
+        if include_base_types and self.base_types is not None:
+            return chain(
+                chain.from_iterable(Typedef.base_type(base_type).values() for base_type in self.base_types if base_type),
+                self._values
+            )
+        return iter(self._values)
 
     def add_value(self, string, doc=None):
+        """
+        TODO
+        """
+
         value = EnumValue(string, doc)
         self._values.append(value)
         return value
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr()
 
     def validate(self, value, unused_mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate the value
         if value not in self.values():
-            raise ValidationError.member_error(self, value, _member)
+            raise _member_error(self, value, _member)
 
         return value
 
 
-# String type
-class _TypeString:
+class TypeString:
+    """
+    TODO
+    """
+
     __slots__ = ()
 
+    #: TODO
     type_name = 'string'
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr(allow_length=True)
 
     def validate(self, value, unused_mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate the value
         if not isinstance(value, str):
-            raise ValidationError.member_error(self, value, _member)
+            raise _member_error(self, value, _member)
 
         return value
 
-TYPE_STRING = _TypeString()
 
+class TypeInt:
+    """
+    TODO
+    """
 
-# Int type
-class _TypeInt:
     __slots__ = ()
 
+    #: TODO
     type_name = 'int'
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr(allow_value=True)
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, int) and not isinstance(value, bool):
@@ -428,31 +727,40 @@ class _TypeInt:
         elif isinstance(value, (float, Decimal)):
             value_x = int(value)
             if value_x != value:
-                raise ValidationError.member_error(self, value, _member)
+                raise _member_error(self, value, _member)
         elif mode == ValidationMode.QUERY_STRING and isinstance(value, str):
             try:
                 value_x = int(value)
             except:
-                raise ValidationError.member_error(self, value, _member)
+                raise _member_error(self, value, _member) from None
         else:
-            raise ValidationError.member_error(self, value, _member)
+            raise _member_error(self, value, _member)
 
         return value if mode in IMMUTABLE_VALIDATION_MODES else value_x
 
-TYPE_INT = _TypeInt()
 
+class TypeFloat:
+    """
+    TODO
+    """
 
-# Float type
-class _TypeFloat:
     __slots__ = ()
 
+    #: TODO
     type_name = 'float'
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr(allow_value=True)
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, float):
@@ -465,21 +773,24 @@ class _TypeFloat:
                 if isnan(value_x) or isinf(value_x):
                     raise ValueError()
             except:
-                raise ValidationError.member_error(self, value, _member)
+                raise _member_error(self, value, _member) from None
         else:
-            raise ValidationError.member_error(self, value, _member)
+            raise _member_error(self, value, _member)
 
         return value if mode in IMMUTABLE_VALIDATION_MODES else value_x
 
-TYPE_FLOAT = _TypeFloat()
 
+class TypeBool:
+    """
+    TODO
+    """
 
-# Bool type
-class _TypeBool:
     __slots__ = ()
 
+    #: TODO
     type_name = 'bool'
 
+    #: TODO
     VALUES = {
         'true': True,
         'false': False
@@ -487,119 +798,178 @@ class _TypeBool:
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr()
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, bool):
             return value
-        elif mode == ValidationMode.QUERY_STRING and isinstance(value, str):
+        if mode == ValidationMode.QUERY_STRING and isinstance(value, str):
             try:
                 return self.VALUES[value]
             except:
-                raise ValidationError.member_error(self, value, _member)
-        else:
-            raise ValidationError.member_error(self, value, _member)
-
-TYPE_BOOL = _TypeBool()
+                raise _member_error(self, value, _member) from None
+        raise _member_error(self, value, _member)
 
 
-# Uuid type
-class _TypeUuid:
+class TypeUuid:
+    """
+    TODO
+    """
+
     __slots__ = ()
 
+    #: TODO
     type_name = 'uuid'
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr()
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, UUID):
             return value
-        elif mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
+        if mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
             try:
                 return UUID(value)
             except:
-                raise ValidationError.member_error(self, value, _member)
-        else:
-            raise ValidationError.member_error(self, value, _member)
-
-TYPE_UUID = _TypeUuid()
+                raise _member_error(self, value, _member) from None
+        raise _member_error(self, value, _member)
 
 
-# Date type
-class _TypeDate:
+class TypeDate:
+    """
+    TODO
+    """
+
     __slots__ = ()
 
+    #: TODO
     type_name = 'date'
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr()
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, date):
             return value
-        elif mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
+        if mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
             try:
-                return parse_iso8601_date(value)
+                return date.fromisoformat(value)
             except:
-                raise ValidationError.member_error(self, value, _member)
-        else:
-            raise ValidationError.member_error(self, value, _member)
-
-TYPE_DATE = _TypeDate()
+                raise _member_error(self, value, _member) from None
+        raise _member_error(self, value, _member)
 
 
-# Datetime type
-class _TypeDatetime:
+class TypeDatetime:
+    """
+    TODO
+    """
+
     __slots__ = ()
 
+    #: TODO
     type_name = 'datetime'
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr()
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()):
+        """
+        TODO
+        """
 
         # Validate and translate the value
         if isinstance(value, datetime):
-            # Set a time zone, if necessary
-            if mode not in IMMUTABLE_VALIDATION_MODES and value.tzinfo is None:
-                return value.replace(tzinfo=TZLOCAL)
             return value
-        elif mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
+        if mode not in IMMUTABLE_VALIDATION_MODES and isinstance(value, str):
             try:
-                return parse_iso8601_datetime(value)
+                return datetime.fromisoformat(value)
             except:
-                raise ValidationError.member_error(self, value, _member)
-        else:
-            raise ValidationError.member_error(self, value, _member)
-
-TYPE_DATETIME = _TypeDatetime()
+                raise _member_error(self, value, _member) from None
+        raise _member_error(self, value, _member)
 
 
-# Object type
-class _TypeObject:
+class TypeObject:
+    """
+    TODO
+    """
+
     __slots__ = ()
 
+    #: TODO
     type_name = 'object'
 
     @staticmethod
     def validate_attr(attr):
+        """
+        TODO
+        """
+
         attr.validate_attr()
 
     def validate(self, value, mode=ValidationMode.DEFAULT, _member=()): # pylint: disable=unused-argument
+        """
+        TODO
+        """
+
         if value is not None:
             return value
-        else:
-            raise ValidationError.member_error(self, value, _member)
+        raise _member_error(self, value, _member)
 
-TYPE_OBJECT = _TypeObject()
+
+#: TODO
+TYPE_BOOL = TypeBool()
+
+#: TODO
+TYPE_DATE = TypeDate()
+
+#: TODO
+TYPE_DATETIME = TypeDatetime()
+
+#: TODO
+TYPE_FLOAT = TypeFloat()
+
+#: TODO
+TYPE_INT = TypeInt()
+
+#: TODO
+TYPE_OBJECT = TypeObject()
+
+#: TODO
+TYPE_STRING = TypeString()
+
+#: TODO
+TYPE_UUID = TypeUuid()
