@@ -1,72 +1,74 @@
 # Licensed under the MIT License
 # https://github.com/craigahobbs/chisel/blob/master/LICENSE
 
+"""
+TODO
+"""
+
+from functools import partial
 from itertools import chain
 import os
 import re
 
-from .model import AttributeValidationError, StructMemberAttributes, TypeArray, \
-    TYPE_BOOL, TYPE_DATE, TYPE_DATETIME, Typedef, TypeDict, TypeEnum, TYPE_INT, \
-    TYPE_FLOAT, TYPE_OBJECT, TYPE_STRING, TypeStruct, TYPE_UUID
+from .model import \
+    ActionModel, AttributeValidationError, StructMemberAttributes, \
+    TYPE_BOOL, TYPE_DATE, TYPE_DATETIME, TYPE_FLOAT, TYPE_INT, TYPE_OBJECT, TYPE_STRING, TYPE_UUID, \
+    TypeArray, TypeDict, TypeEnum, TypeStruct, Typedef
 
 
-# Action model
-class ActionModel:
-    __slots__ = ('name', 'input_type', 'output_type', 'error_type', 'doc', 'doc_group')
-
-    def __init__(self, name, doc=None, doc_group=None):
-        self.name = name
-        self.input_type = TypeStruct(type_name=name + '_input')
-        self.output_type = TypeStruct(type_name=name + '_output')
-        self.error_type = TypeEnum(type_name=name + '_error')
-        self.doc = [] if doc is None else doc
-        self.doc_group = doc_group
-
-
-# Spec parser exception
 class SpecParserError(Exception):
+    """
+    TODO
+    """
+
     __slots__ = ('errors',)
 
     def __init__(self, errors):
         super().__init__('\n'.join(errors))
+
+        #: TODO
         self.errors = errors
 
 
 # Spec language regex
-_RE_PART_ID = r'(?:[A-Za-z]\w*)'
-_RE_PART_ATTR_GROUP = r'(?:(?P<op><=|<|>|>=|==)\s*(?P<opnum>-?\d+(?:\.\d+)?)' \
-                      r'|(?P<ltype>len)\s*(?P<lop><=|<|>|>=|==)\s*(?P<lopnum>\d+))'
-_RE_PART_ATTR = re.sub(r'\(\?P<[^>]+>', r'(?:', _RE_PART_ATTR_GROUP)
-_RE_PART_ATTRS = r'(?:' + _RE_PART_ATTR + r'(?:\s*,\s*' + _RE_PART_ATTR + r')*)'
-_RE_ATTR_GROUP = re.compile(_RE_PART_ATTR_GROUP)
-_RE_FIND_ATTRS = re.compile(_RE_PART_ATTR + r'(?:\s*,\s*|\s*\Z)')
-_RE_LINE_CONT = re.compile(r'\\s*$')
-_RE_COMMENT = re.compile(r'^\s*(?:#-.*|#(?P<doc>.*))?$')
-_RE_GROUP = re.compile(r'^group(?:\s+"(?P<group>.+?)")?\s*$')
-_RE_ACTION = re.compile(r'^action\s+(?P<id>' + _RE_PART_ID + r')')
-_RE_PART_BASE_IDS = r'(?:\s*\(\s*(?P<base_ids>' + _RE_PART_ID + r'(?:\s*,\s*' + _RE_PART_ID + r')*)\s*\)\s*)'
-_RE_BASE_IDS_SPLIT = re.compile(r'\s*,\s*')
-_RE_DEFINITION = re.compile(r'^(?P<type>action|struct|union|enum)\s+(?P<id>' + _RE_PART_ID + r')' + _RE_PART_BASE_IDS + r'?\s*$')
-_RE_SECTION = re.compile(r'^\s+(?P<type>input|output|errors)' + _RE_PART_BASE_IDS + r'?\s*$')
-_RE_PART_TYPEDEF = r'(?P<type>' + _RE_PART_ID + r')' \
-                   r'(?:\s*\(\s*(?P<attrs>' + _RE_PART_ATTRS + r')\s*\))?' \
-                   r'(?:' \
-                   r'(?:\s*\[\s*(?P<array>' + _RE_PART_ATTRS + r'?)\s*\])?' \
-                   r'|' \
-                   r'(?:' \
-                   r'\s*:\s*(?P<dictValueType>' + _RE_PART_ID + r')' \
-                   r'(?:\s*\(\s*(?P<dictValueAttrs>' + _RE_PART_ATTRS + r')\s*\))?' \
-                   r')?' \
-                   r'(?:\s*\{\s*(?P<dict>' + _RE_PART_ATTRS + r'?)\s*\})?' \
-                   r')' \
-                   r'\s+(?P<id>' + _RE_PART_ID + r')'
-_RE_TYPEDEF = re.compile(r'^typedef\s+' + _RE_PART_TYPEDEF + r'\s*$')
-_RE_MEMBER = re.compile(r'^\s+(?P<optional>optional\s+)?(?P<nullable>nullable\s+)?' + _RE_PART_TYPEDEF + r'\s*$')
-_RE_VALUE = re.compile(r'^\s+"?(?P<id>(?<!")' + _RE_PART_ID + r'(?!")|(?<=").*?(?="))"?\s*$')
+RE_PART_ID = r'(?:[A-Za-z]\w*)'
+RE_PART_ATTR_GROUP = \
+    r'(?:(?P<op><=|<|>|>=|==)\s*(?P<opnum>-?\d+(?:\.\d+)?)' \
+    r'|(?P<ltype>len)\s*(?P<lop><=|<|>|>=|==)\s*(?P<lopnum>\d+))'
+RE_PART_ATTR = re.sub(r'\(\?P<[^>]+>', r'(?:', RE_PART_ATTR_GROUP)
+RE_PART_ATTRS = r'(?:' + RE_PART_ATTR + r'(?:\s*,\s*' + RE_PART_ATTR + r')*)'
+RE_ATTR_GROUP = re.compile(RE_PART_ATTR_GROUP)
+RE_FIND_ATTRS = re.compile(RE_PART_ATTR + r'(?:\s*,\s*|\s*\Z)')
+RE_LINE_CONT = re.compile(r'\\s*$')
+RE_COMMENT = re.compile(r'^\s*(?:#-.*|#(?P<doc>.*))?$')
+RE_GROUP = re.compile(r'^group(?:\s+"(?P<group>.+?)")?\s*$')
+RE_ACTION = re.compile(r'^action\s+(?P<id>' + RE_PART_ID + r')')
+RE_PART_BASE_IDS = r'(?:\s*\(\s*(?P<base_ids>' + RE_PART_ID + r'(?:\s*,\s*' + RE_PART_ID + r')*)\s*\)\s*)'
+RE_BASE_IDS_SPLIT = re.compile(r'\s*,\s*')
+RE_DEFINITION = re.compile(r'^(?P<type>struct|union|enum)\s+(?P<id>' + RE_PART_ID + r')' + RE_PART_BASE_IDS + r'?\s*$')
+RE_SECTION = re.compile(r'^\s+(?P<type>path|query|input|output|errors)' + RE_PART_BASE_IDS + r'?\s*$')
+RE_SECTION_PLAIN = re.compile(r'^\s+(?P<type>url)\s*$')
+RE_PART_TYPEDEF = \
+    r'(?P<type>' + RE_PART_ID + r')' \
+    r'(?:\s*\(\s*(?P<attrs>' + RE_PART_ATTRS + r')\s*\))?' \
+    r'(?:' \
+    r'(?:\s*\[\s*(?P<array>' + RE_PART_ATTRS + r'?)\s*\])?' \
+    r'|' \
+    r'(?:' \
+    r'\s*:\s*(?P<dictValueType>' + RE_PART_ID + r')' \
+    r'(?:\s*\(\s*(?P<dictValueAttrs>' + RE_PART_ATTRS + r')\s*\))?' \
+    r')?' \
+    r'(?:\s*\{\s*(?P<dict>' + RE_PART_ATTRS + r'?)\s*\})?' \
+    r')' \
+    r'\s+(?P<id>' + RE_PART_ID + r')'
+RE_TYPEDEF = re.compile(r'^typedef\s+' + RE_PART_TYPEDEF + r'\s*$')
+RE_MEMBER = re.compile(r'^\s+(?P<optional>optional\s+)?(?P<nullable>nullable\s+)?' + RE_PART_TYPEDEF + r'\s*$')
+RE_VALUE = re.compile(r'^\s+"?(?P<id>(?<!")' + RE_PART_ID + r'(?!")|(?<=").*?(?="))"?\s*$')
+RE_URL = re.compile(r'^\s+(?P<method>[A-Za-z]+|\*)(?:\s+(?P<path>/[^\s]*))?')
 
 
 # Built-in types
-_TYPES = {
+TYPES = {
     'bool': TYPE_BOOL,
     'date': TYPE_DATE,
     'datetime': TYPE_DATETIME,
@@ -80,6 +82,10 @@ _TYPES = {
 
 # Specification language parser class
 class SpecParser:
+    """
+    TODO
+    """
+
     __slots__ = (
         'types',
         'actions',
@@ -89,6 +95,7 @@ class SpecParser:
         '_parse_linenum',
         '_action',
         '_action_sections',
+        '_urls',
         '_type',
         '_doc',
         '_doc_group',
@@ -97,13 +104,21 @@ class SpecParser:
     )
 
     def __init__(self, spec=None):
+
+        #: TODO
         self.types = {}
+
+        #: TODO
         self.actions = {}
+
+        #: TODO
         self.errors = []
+
         self._typerefs = []
         self._finalize_checks = []
         self._action = None
         self._action_sections = None
+        self._urls = None
         self._type = None
         self._doc = None
         self._doc_group = None
@@ -115,9 +130,13 @@ class SpecParser:
 
     # Parse a specification from an iterator of spec lines (e.g an input stream)
     def parse(self, lines, filename='', finalize=True):
+        """
+        TODO
+        """
 
         # Set the parser state
         self._action = None
+        self._urls = None
         self._type = None
         self._doc = []
         self._doc_group = None
@@ -132,10 +151,17 @@ class SpecParser:
 
     # Parse a specification string
     def parse_string(self, spec, filename='', finalize=True):
+        """
+        TODO
+        """
+
         self.parse(spec.splitlines(), finalize=finalize, filename=filename)
 
     # Finalize parsing (must call after calling parse one or more times - can be repeated)
     def finalize(self):
+        """
+        TODO
+        """
 
         # Fixup type refs
         for typeref in self._typerefs:
@@ -152,6 +178,10 @@ class SpecParser:
             raise SpecParserError(self.errors)
 
     def load(self, path, spec_ext='.chsl', finalize=True):
+        """
+        TODO
+        """
+
         if os.path.isdir(path):
             for dirpath, _, filenames in os.walk(path):
                 for filename in filenames:
@@ -173,7 +203,7 @@ class SpecParser:
         linenum = self._parse_linenum
 
         def set_type(error):
-            type_ = _TYPES.get(type_name) or self.types.get(type_name)
+            type_ = TYPES.get(type_name) or self.types.get(type_name)
             if type_ is not None:
                 error_count = len(self.errors)
                 if validate_fn is not None:
@@ -201,31 +231,13 @@ class SpecParser:
         if not isinstance(base_type_base, TypeStruct) or base_type_base.union != struct_type.union:
             if def_type is None:
                 def_type = 'union' if struct_type.union else 'struct'
-            self._error('Invalid ' + def_type + " base type '" + base_type.type_name + "'", filename, linenum)
+            self._error(f"Invalid {def_type} base type '{base_type.type_name}'", filename, linenum)
 
     def _validate_enum_base_type(self, unused_enum_type, base_type, filename, linenum, def_type=None):
         if not isinstance(Typedef.base_type(base_type), TypeEnum):
             if def_type is None:
                 def_type = 'enum'
-            self._error('Invalid ' + def_type + " base type '" + base_type.type_name + "'", filename, linenum)
-
-    def _validate_input_base_type(self, *args):
-        self._validate_struct_base_type(*args, def_type='action input')
-
-    def _validate_output_base_type(self, *args):
-        self._validate_struct_base_type(*args, def_type='action output')
-
-    def _validate_errors_base_type(self, *args):
-        self._validate_enum_base_type(*args, def_type='action errors')
-
-    def _set_finalize(self, type_, finalize_fn):
-        filename = self._parse_filename
-        linenum = self._parse_linenum
-
-        def finalize_check():
-            finalize_fn(type_, filename, linenum)
-
-        self._finalize_checks.append(finalize_check)
+            self._error(f"Invalid {def_type} base type '{base_type.type_name}'", filename, linenum)
 
     def _finalize_circular_base_type(self, type_, filename, linenum):
         is_circular = False
@@ -244,15 +256,19 @@ class SpecParser:
                 self._error("Circular base type detected for type '" + base_type_name + "'", filename, linenum)
         return is_circular
 
-    def _finalize_struct_base_type(self, struct_type, filename, linenum):
+    def _finalize_struct_base_type(self, action, struct_type, filename, linenum):
         if struct_type.base_types is not None and not self._finalize_circular_base_type(struct_type, filename, linenum):
             members = {}
-            for member in struct_type.members():
+            if action is None or struct_type is action.output_type:
+                struct_members = struct_type.members
+            else:
+                struct_members = action.input_members
+            for member in struct_members():
                 member_count = members[member.name] = members.get(member.name, 0) + 1
                 if member_count == 2:
                     self._error("Redefinition of member '" + member.name + "' from base type", filename, linenum)
 
-    def _finalize_enum_base_type(self, enum_type, filename, linenum):
+    def _finalize_enum_base_type(self, unused_action, enum_type, filename, linenum):
         if enum_type.base_types is not None and not self._finalize_circular_base_type(enum_type, filename, linenum):
             values = {}
             for value in enum_type.values():
@@ -262,7 +278,7 @@ class SpecParser:
 
     # Record an error
     def _error(self, msg, filename=None, linenum=None):
-        self.errors.append('{0}:{1}: error: {2}'.format(filename or self._parse_filename, linenum or self._parse_linenum, msg))
+        self.errors.append(f'{filename or self._parse_filename}:{linenum or self._parse_linenum}: error: {msg}')
 
     # Validate a type's attributes
     def _validate_attr(self, type_, attr, filename=None, linenum=None):
@@ -275,41 +291,44 @@ class SpecParser:
     # Parse an attributes string
     @classmethod
     def _parse_attr(cls, attrs_string):
-        attr = None
+        has_attrs = False
+        op_lt, op_lte, op_gt, op_gte, op_eq, op_len_lt, op_len_lte, op_len_gt, op_len_gte, op_len_eq = \
+            None, None, None, None, None, None, None, None, None, None
         if attrs_string is not None:
-            for attr_string in _RE_FIND_ATTRS.findall(attrs_string):
-                match_attr = _RE_ATTR_GROUP.match(attr_string)
+            for attr_string in RE_FIND_ATTRS.findall(attrs_string):
+                has_attrs = True
+                match_attr = RE_ATTR_GROUP.match(attr_string)
                 attr_op = match_attr.group('op')
                 attr_length_op = match_attr.group('lop') if attr_op is None else None
 
-                if attr is None:
-                    attr = StructMemberAttributes()
-
                 if attr_op is not None:
                     attr_value = float(match_attr.group('opnum'))
                     if attr_op == '<':
-                        attr.op_lt = attr_value
+                        op_lt = attr_value
                     elif attr_op == '<=':
-                        attr.op_lte = attr_value
+                        op_lte = attr_value
                     elif attr_op == '>':
-                        attr.op_gt = attr_value
+                        op_gt = attr_value
                     elif attr_op == '>=':
-                        attr.op_gte = attr_value
+                        op_gte = attr_value
                     else:  # ==
-                        attr.op_eq = attr_value
+                        op_eq = attr_value
                 else:  # attr_length_op is not None:
                     attr_value = int(match_attr.group('lopnum'))
                     if attr_length_op == '<':
-                        attr.op_len_lt = attr_value
+                        op_len_lt = attr_value
                     elif attr_length_op == '<=':
-                        attr.op_len_lte = attr_value
+                        op_len_lte = attr_value
                     elif attr_length_op == '>':
-                        attr.op_len_gt = attr_value
+                        op_len_gt = attr_value
                     elif attr_length_op == '>=':
-                        attr.op_len_gte = attr_value
+                        op_len_gte = attr_value
                     else:  # ==
-                        attr.op_len_eq = attr_value
-        return attr
+                        op_len_eq = attr_value
+
+        if not has_attrs:
+            return None
+        return StructMemberAttributes(op_eq, op_lt, op_lte, op_gt, op_gte, op_len_eq, op_len_lt, op_len_lte, op_len_gt, op_len_gte)
 
     # Construct typedef parts
     def _parse_typedef(self, parent, parent_type_attr, parent_attr_attr, match_typedef):
@@ -369,38 +388,56 @@ class SpecParser:
             self._parse_linenum += 1
 
             # Line continuation?
-            line_part_no_continuation = _RE_LINE_CONT.sub('', line_part)
+            line_part_no_continuation = RE_LINE_CONT.sub('', line_part)
             if line_continuation or line_part_no_continuation is not line_part:
                 line_continuation.append(line_part_no_continuation)
             if line_part_no_continuation is not line_part:
                 continue
-            elif line_continuation:
+            if line_continuation:
                 line = ''.join(line_continuation)
                 del line_continuation[:]
             else:
                 line = line_part
 
+            # Match syntax
+            match_name, match = 'comment', RE_COMMENT.search(line)
+            if match is None:
+                match_name, match = 'group', RE_GROUP.search(line)
+            if match is None:
+                match_name, match = 'action', RE_ACTION.search(line)
+            if match is None:
+                match_name, match = 'definition', RE_DEFINITION.search(line)
+            if match is None and self._action is not None:
+                match_name, match = 'section', RE_SECTION.search(line)
+            if match is None and self._action is not None:
+                match_name, match = 'section_plain', RE_SECTION_PLAIN.search(line)
+            if match is None and isinstance(self._type, TypeEnum):
+                match_name, match = 'value', RE_VALUE.search(line)
+            if match is None and isinstance(self._type, TypeStruct):
+                match_name, match = 'member', RE_MEMBER.search(line)
+            if match is None and self._urls is not None:
+                match_name, match = 'url', RE_URL.search(line)
+            if match is None:
+                match_name, match = 'typedef', RE_TYPEDEF.search(line)
+            if match is None:
+                match_name = None
+
             # Comment?
-            match = _RE_COMMENT.search(line)
-            if match:
+            if match_name == 'comment':
                 doc_string = match.group('doc')
                 if doc_string is not None:
-                    self._doc.append(doc_string.strip())
-
-                continue
+                    self._doc.append(doc_string)
 
             # Documentation group?
-            match = _RE_GROUP.search(line)
-            if match:
-                self._doc_group = match.group('group')
-                if self._doc_group is not None:
-                    self._doc_group = self._doc_group.strip()
-
-                continue
+            elif match_name == 'group':
+                doc_group = match.group('group')
+                if doc_group is not None:
+                    self._doc_group = doc_group.strip()
+                else:
+                    self._doc_group = None
 
             # Action?
-            match = _RE_ACTION.search(line)
-            if match:
+            elif match_name == 'action':
                 action_id = match.group('id')
 
                 # Action already defined?
@@ -410,36 +447,37 @@ class SpecParser:
                 # Create the new action
                 self._action = ActionModel(action_id, doc=self._doc, doc_group=self._doc_group)
                 self._action_sections = set()
+                self._urls = None
                 self._type = None
                 self._doc = []
                 self.actions[self._action.name] = self._action
 
-                continue
-
             # Definition?
-            match = _RE_DEFINITION.search(line)
-            if match:
+            elif match_name == 'definition':
                 definition_string = match.group('type')
                 definition_id = match.group('id')
                 definition_base_ids = match.group('base_ids')
                 if definition_base_ids is not None:
-                    definition_base_ids = _RE_BASE_IDS_SPLIT.split(definition_base_ids)
+                    definition_base_ids = RE_BASE_IDS_SPLIT.split(definition_base_ids)
 
                 # Struct definition
                 if definition_string in ('struct', 'union'):
 
                     # Type already defined?
-                    if definition_id in _TYPES or definition_id in self.types:
+                    if definition_id in TYPES or definition_id in self.types:
                         self._error("Redefinition of type '" + definition_id + "'")
 
                     # Create the new struct type
                     self._action = None
+                    self._urls = None
                     self._type = TypeStruct(type_name=definition_id, union=(definition_string == 'union'), doc=self._doc)
                     if definition_base_ids is not None:
                         self._type.base_types = [None for _ in definition_base_ids]
                         for base_index, base_id in enumerate(definition_base_ids):
                             self._set_type(self._type, self._type.base_types, base_index, base_id, None, self._validate_struct_base_type)
-                        self._set_finalize(self._type, self._finalize_struct_base_type)
+                        self._finalize_checks.append(
+                            partial(self._finalize_struct_base_type, None, self._type, self._parse_filename, self._parse_linenum)
+                        )
                     self._doc = []
                     self.types[self._type.type_name] = self._type
 
@@ -447,82 +485,82 @@ class SpecParser:
                 else:  # definition_string == 'enum':
 
                     # Type already defined?
-                    if definition_id in _TYPES or definition_id in self.types:
+                    if definition_id in TYPES or definition_id in self.types:
                         self._error("Redefinition of type '" + definition_id + "'")
 
                     # Create the new enum type
                     self._action = None
+                    self._urls = None
                     self._type = TypeEnum(type_name=definition_id, doc=self._doc)
                     if definition_base_ids is not None:
                         self._type.base_types = [None for _ in definition_base_ids]
                         for base_index, base_id in enumerate(definition_base_ids):
                             self._set_type(self._type, self._type.base_types, base_index, base_id, None, self._validate_enum_base_type)
-                        self._set_finalize(self._type, self._finalize_enum_base_type)
+                        self._finalize_checks.append(
+                            partial(self._finalize_enum_base_type, None, self._type, self._parse_filename, self._parse_linenum)
+                        )
                     self._doc = []
                     self.types[self._type.type_name] = self._type
 
-                continue # definition
-
             # Section?
-            match = _RE_SECTION.search(line)
-            if match:
+            elif match_name == 'section':
                 section_string = match.group('type')
                 section_base_ids = match.group('base_ids')
                 if section_base_ids is not None:
-                    section_base_ids = _RE_BASE_IDS_SPLIT.split(section_base_ids)
-
-                # Not in an action scope?
-                if self._action is None:
-                    self._error('Action section outside of action scope')
-                    continue
+                    section_base_ids = RE_BASE_IDS_SPLIT.split(section_base_ids)
 
                 # Action section redefinition?
                 if section_string in self._action_sections:
-                    self._error('Redefinition of action ' + section_string)
-                    self._type = None
-                    continue
+                    self._error(f'Redefinition of action {section_string}')
                 self._action_sections.add(section_string)
 
                 # Set the action section type
-                if section_string == 'input':
+                self._urls = None
+                if section_string == 'path':
+                    self._type = self._action.path_type
+                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
+                elif section_string == 'query':
+                    self._type = self._action.query_type
+                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
+                elif section_string == 'input':
                     self._type = self._action.input_type
-                    if section_base_ids is not None:
-                        self._type.base_types = [None for _ in section_base_ids]
-                        for base_index, base_id in enumerate(section_base_ids):
-                            self._set_type(self._action.input_type, self._action.input_type.base_types, base_index, base_id,
-                                           None, self._validate_input_base_type)
-                        self._set_finalize(self._type, self._finalize_struct_base_type)
-
+                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
                 elif section_string == 'output':
                     self._type = self._action.output_type
-                    if section_base_ids is not None:
-                        self._type.base_types = [None for _ in section_base_ids]
-                        for base_index, base_id in enumerate(section_base_ids):
-                            self._set_type(self._action.output_type, self._action.output_type.base_types, base_index, base_id,
-                                           None, self._validate_output_base_type)
-                        self._set_finalize(self._type, self._finalize_struct_base_type)
-
-                else:  # section_string == 'errors':
+                    validate_base_type_fn, finalize_base_type_fn = self._validate_struct_base_type, self._finalize_struct_base_type
+                else: # section_string == 'errors':
                     self._type = self._action.error_type
-                    if section_base_ids is not None:
-                        self._type.base_types = [None for _ in section_base_ids]
-                        for base_index, base_id in enumerate(section_base_ids):
-                            self._set_type(self._action.error_type, self._action.error_type.base_types, base_index, base_id,
-                                           None, self._validate_errors_base_type)
-                        self._set_finalize(self._type, self._finalize_enum_base_type)
+                    validate_base_type_fn, finalize_base_type_fn = self._validate_enum_base_type, self._finalize_enum_base_type
 
-                continue # section
+                if section_base_ids is not None:
+                    self._type.base_types = [None for _ in section_base_ids]
+                    for base_index, base_id in enumerate(section_base_ids):
+                        self._set_type(
+                            self._type, self._type.base_types, base_index, base_id, None,
+                            partial(validate_base_type_fn, def_type=f'action {section_string}')
+                        )
+                    self._finalize_checks.append(
+                        partial(finalize_base_type_fn, self._action, self._type, self._parse_filename, self._parse_linenum)
+                    )
+
+            # Plain section?
+            elif match_name == 'section_plain':
+                section_string = match.group('type')
+
+                # Action section redefinition?
+                if section_string in self._action_sections:
+                    self._error(f'Redefinition of action {section_string}')
+                self._action_sections.add(section_string)
+
+                # Set the action section data
+                assert section_string == 'url'
+                self._urls = self._action.urls
+                self._type = None
 
             # Enum value?
-            match = _RE_VALUE.search(line)
-            if match:
+            elif match_name == 'value':
                 value_string = match.group('id')
 
-                # Not in an enum scope?
-                if not isinstance(self._type, TypeEnum):
-                    self._error('Enumeration value outside of enum scope')
-                    continue
-
                 # Redefinition of enum value?
                 if value_string in self._type.values(include_base_types=False):
                     self._error("Redefinition of enumeration value '" + value_string + "'")
@@ -531,22 +569,18 @@ class SpecParser:
                 self._type.add_value(value_string, doc=self._doc)
                 self._doc = []
 
-                continue
-
             # Struct member?
-            match = _RE_MEMBER.search(line)
-            if match:
+            elif match_name == 'member':
                 optional = match.group('optional') is not None
                 nullable = match.group('nullable') is not None
                 member_name = match.group('id')
 
-                # Not in a struct scope?
-                if not isinstance(self._type, TypeStruct):
-                    self._error('Member definition outside of struct scope')
-                    continue
-
                 # Member name already defined?
-                if any(member.name == member_name for member in self._type.members(include_base_types=False)):
+                if self._action is None or self._type is self._action.output_type:
+                    struct_members = self._type.members
+                else:
+                    struct_members = self._action.input_members
+                if member_name in (mem.name for mem in struct_members(include_base_types=False)):
                     self._error("Redefinition of member '" + member_name + "'")
 
                 # Create the member
@@ -554,15 +588,26 @@ class SpecParser:
                 self._parse_typedef(member, 'type', 'attr', match)
                 self._doc = []
 
-                continue
+            # URL?
+            elif match_name == 'url':
+                method = match.group('method')
+                path = match.group('path')
+                url = (method if method != '*' else None, path)
+
+                # Duplicate URL?
+                if url in self._urls:
+                    self._error(f'Duplicate URL: {method} {"" if path is None else path}')
+
+                # Add the URL
+                self._urls.append(url)
+                self._doc = []
 
             # Typedef?
-            match = _RE_TYPEDEF.search(line)
-            if match:
+            elif match_name == 'typedef':
                 typedef_name = match.group('id')
 
                 # Type already defined?
-                if typedef_name in _TYPES or typedef_name in self.types:
+                if typedef_name in TYPES or typedef_name in self.types:
                     self._error("Redefinition of type '" + typedef_name + "'")
 
                 # Create the typedef
@@ -572,10 +617,10 @@ class SpecParser:
 
                 # Reset current action/type
                 self._action = None
+                self._urls = None
                 self._type = None
                 self._doc = []
 
-                continue
-
             # Unrecognized line syntax
-            self._error('Syntax error')
+            else:
+                self._error('Syntax error')
